@@ -106,9 +106,9 @@ def test_min_informative_str():
 
     mis = min_informative_str(G).replace("\t", "        ")
 
-    reference = """A. Elemwise{add,no_inplace}
+    reference = """A. Add
  B. C
- C. Elemwise{add,no_inplace}
+ C. Add
   D. D
   E. E"""
 
@@ -144,13 +144,13 @@ def test_debugprint():
     s = s.getvalue()
     reference = dedent(
         r"""
-        Elemwise{add,no_inplace} [id 0]
-         |Elemwise{add,no_inplace} [id 1] 'C'
-         | |A [id 2]
-         | |B [id 3]
-         |Elemwise{add,no_inplace} [id 4]
-           |D [id 5]
-           |E [id 6]
+        Add [id 0]
+         ├─ Add [id 1] 'C'
+         │  ├─ A [id 2]
+         │  └─ B [id 3]
+         └─ Add [id 4]
+            ├─ D [id 5]
+            └─ E [id 6]
         """
     ).lstrip()
 
@@ -162,13 +162,13 @@ def test_debugprint():
     # The additional white space are needed!
     reference = dedent(
         r"""
-        Elemwise{add,no_inplace} [id A]
-         |Elemwise{add,no_inplace} [id B] 'C'
-         | |A [id C]
-         | |B [id D]
-         |Elemwise{add,no_inplace} [id E]
-           |D [id F]
-           |E [id G]
+        Add [id A]
+         ├─ Add [id B] 'C'
+         │  ├─ A [id C]
+         │  └─ B [id D]
+         └─ Add [id E]
+            ├─ D [id F]
+            └─ E [id G]
         """
     ).lstrip()
 
@@ -180,11 +180,12 @@ def test_debugprint():
     # The additional white space are needed!
     reference = dedent(
         r"""
-        Elemwise{add,no_inplace} [id A]
-         |Elemwise{add,no_inplace} [id B] 'C'
-         |Elemwise{add,no_inplace} [id C]
-           |D [id D]
-           |E [id E]
+        Add [id A]
+         ├─ Add [id B] 'C'
+         │  └─ ···
+         └─ Add [id C]
+            ├─ D [id D]
+            └─ E [id E]
         """
     ).lstrip()
 
@@ -195,13 +196,13 @@ def test_debugprint():
     s = s.getvalue()
     reference = dedent(
         r"""
-        Elemwise{add,no_inplace}
-         |Elemwise{add,no_inplace} 'C'
-         | |A
-         | |B
-         |Elemwise{add,no_inplace}
-           |D
-           |E
+        Add
+         ├─ Add 'C'
+         │  ├─ A
+         │  └─ B
+         └─ Add
+            ├─ D
+            └─ E
         """
     ).lstrip()
 
@@ -212,11 +213,11 @@ def test_debugprint():
     s = s.getvalue()
     reference = dedent(
         r"""
-        Elemwise{add,no_inplace} 0 [None]
-         |A [None]
-         |B [None]
-         |D [None]
-         |E [None]
+        Add 0 [None]
+         ├─ A [None]
+         ├─ B [None]
+         ├─ D [None]
+         └─ E [None]
         """
     ).lstrip()
 
@@ -230,11 +231,11 @@ def test_debugprint():
     s = s.getvalue()
     reference = dedent(
         r"""
-        Elemwise{add,no_inplace} 0 [None]
-         |A [None]
-         |B [None]
-         |D [None]
-         |E [None]
+        Add 0 [None]
+         ├─ A [None]
+         ├─ B [None]
+         ├─ D [None]
+         └─ E [None]
         """
     ).lstrip()
 
@@ -248,11 +249,11 @@ def test_debugprint():
     s = s.getvalue()
     reference = dedent(
         r"""
-        Elemwise{add,no_inplace} 0 [None]
-         |A [None]
-         |B [None]
-         |D [None]
-         |E [None]
+        Add 0 [None]
+         ├─ A [None]
+         ├─ B [None]
+         ├─ D [None]
+         └─ E [None]
         """
     ).lstrip()
 
@@ -273,27 +274,27 @@ def test_debugprint():
     s = s.getvalue()
     exp_res = dedent(
         r"""
-        Elemwise{Composite} 4
-         |InplaceDimShuffle{x,0} v={0: [0]} 3
-         | |CGemv{inplace} d={0: [0]} 2
-         |   |AllocEmpty{dtype='float64'} 1
-         |   | |Shape_i{0} 0
-         |   |   |B
-         |   |TensorConstant{1.0}
-         |   |B
-         |   |<TensorType(float64, (?,))>
-         |   |TensorConstant{0.0}
-         |D
-         |A
+        Composite{(i2 + (i0 - i1))} 4
+         ├─ ExpandDims{axis=0} v={0: [0]} 3
+         │  └─ CGemv{inplace} d={0: [0]} 2
+         │     ├─ AllocEmpty{dtype='float64'} 1
+         │     │  └─ Shape_i{0} 0
+         │     │     └─ B
+         │     ├─ 1.0
+         │     ├─ B
+         │     ├─ <Vector(float64, shape=(?,))>
+         │     └─ 0.0
+         ├─ D
+         └─ A
 
         Inner graphs:
 
-        Elemwise{Composite}
-         >add
-         > |<float64>
-         > |sub
-         >   |<float64>
-         >   |<float64>
+        Composite{(i2 + (i0 - i1))}
+         ← add 'o0'
+            ├─ i2
+            └─ sub
+               ├─ i0
+               └─ i1
         """
     ).lstrip()
 
@@ -313,11 +314,11 @@ def test_debugprint_id_type():
     debugprint(e_at, id_type="auto", file=s)
     s = s.getvalue()
 
-    exp_res = f"""Elemwise{{add,no_inplace}} [id {e_at.auto_name}]
- |dot [id {d_at.auto_name}]
- | |<TensorType(float64, (?, ?))> [id {b_at.auto_name}]
- | |<TensorType(float64, (?,))> [id {a_at.auto_name}]
- |<TensorType(float64, (?,))> [id {a_at.auto_name}]
+    exp_res = f"""Add [id {e_at.auto_name}]
+ ├─ dot [id {d_at.auto_name}]
+ │  ├─ <Matrix(float64, shape=(?, ?))> [id {b_at.auto_name}]
+ │  └─ <Vector(float64, shape=(?,))> [id {a_at.auto_name}]
+ └─ <Vector(float64, shape=(?,))> [id {a_at.auto_name}]
     """
 
     assert [l.strip() for l in s.split("\n")] == [
@@ -328,7 +329,7 @@ def test_debugprint_id_type():
 def test_pprint():
     x = dvector()
     y = x[1]
-    assert pp(y) == "<TensorType(float64, (?,))>[1]"
+    assert pp(y) == "<Vector(float64, shape=(?,))>[1]"
 
 
 def test_debugprint_inner_graph():
@@ -351,15 +352,15 @@ def test_debugprint_inner_graph():
     lines = output_str.split("\n")
 
     exp_res = """MyInnerGraphOp [id A]
- |3 [id B]
- |4 [id C]
+ ├─ 3 [id B]
+ └─ 4 [id C]
 
 Inner graphs:
 
 MyInnerGraphOp [id A]
- >op2 [id D] 'igo1'
- > |*0-<MyType()> [id E]
- > |*1-<MyType()> [id F]
+ ← op2 [id D] 'igo1'
+    ├─ *0-<MyType()> [id E]
+    └─ *1-<MyType()> [id F]
     """
 
     for exp_line, res_line in zip(exp_res.split("\n"), lines):
@@ -375,19 +376,19 @@ MyInnerGraphOp [id A]
     lines = output_str.split("\n")
 
     exp_res = """MyInnerGraphOp [id A]
- |5 [id B]
+ └─ 5 [id B]
 
 Inner graphs:
 
 MyInnerGraphOp [id A]
- >MyInnerGraphOp [id C]
- > |*0-<MyType()> [id D]
- > |*1-<MyType()> [id E]
+ ← MyInnerGraphOp [id C]
+    ├─ *0-<MyType()> [id D]
+    └─ *1-<MyType()> [id E]
 
 MyInnerGraphOp [id C]
- >op2 [id F] 'igo1'
- > |*0-<MyType()> [id D]
- > |*1-<MyType()> [id E]
+ ← op2 [id F] 'igo1'
+    ├─ *0-<MyType()> [id D]
+    └─ *1-<MyType()> [id E]
     """
 
     for exp_line, res_line in zip(exp_res.split("\n"), lines):
